package dev.fyloz.colorrecipesexplorer.logic

import dev.fyloz.colorrecipesexplorer.config.annotations.LogicComponent
import dev.fyloz.colorrecipesexplorer.dtos.MaterialDto
import dev.fyloz.colorrecipesexplorer.dtos.MaterialTypeDto
import dev.fyloz.colorrecipesexplorer.dtos.MixTypeDto
import dev.fyloz.colorrecipesexplorer.model.MixType
import dev.fyloz.colorrecipesexplorer.service.MixTypeService
import org.springframework.transaction.annotation.Transactional

interface MixTypeLogic : Logic<MixTypeDto, MixTypeService> {
    /** Returns a [MixType] for the given [name] and [materialType]. If this mix type does not already exist, it will be created. */
    fun getOrCreateForNameAndMaterialType(name: String, materialType: MaterialTypeDto): MixTypeDto

    /** Returns a new and persisted [MixType] with the given [name] and [materialType]. */
    fun saveForNameAndMaterialType(name: String, materialType: MaterialTypeDto): MixTypeDto

    /** Returns the given [mixType] updated with the given [name] and [materialType]. */
    fun updateForNameAndMaterialType(mixType: MixTypeDto, name: String, materialType: MaterialTypeDto): MixTypeDto
}

@LogicComponent
class DefaultMixTypeLogic(service: MixTypeService, private val materialLogic: MaterialLogic) :
    BaseLogic<MixTypeDto, MixTypeService>(service, MixType::class.simpleName!!), MixTypeLogic {
    override fun getOrCreateForNameAndMaterialType(name: String, materialType: MaterialTypeDto) =
        service.getByNameAndMaterialType(name, materialType.id) ?: saveForNameAndMaterialType(name, materialType)

    @Transactional
    override fun saveForNameAndMaterialType(name: String, materialType: MaterialTypeDto): MixTypeDto {
        val material = materialLogic.save(
            MaterialDto(
                name = name,
                inventoryQuantity = Float.MIN_VALUE,
                isMixType = true,
                materialType = materialType
            )
        )

        return save(MixTypeDto(name = name, material = material))
    }

    override fun updateForNameAndMaterialType(
        mixType: MixTypeDto,
        name: String,
        materialType: MaterialTypeDto
    ): MixTypeDto {
        val material = materialLogic.update(mixType.material.copy(name = name, materialType = materialType))
        return update(mixType.copy(name = name, material = material))
    }

    override fun deleteById(id: Long) {
        if (service.isUsedByMixes(id)) {
            throw cannotDeleteException("Cannot delete the mix type with the id '$id' because one or more mixes depends on it")
        }

        super.deleteById(id)
    }
}