package dev.fyloz.colorrecipesexplorer.logic

import dev.fyloz.colorrecipesexplorer.dtos.MaterialDto
import dev.fyloz.colorrecipesexplorer.dtos.MaterialTypeDto
import dev.fyloz.colorrecipesexplorer.dtos.MixTypeDto
import dev.fyloz.colorrecipesexplorer.exception.CannotDeleteException
import dev.fyloz.colorrecipesexplorer.service.MixTypeService
import io.mockk.*
import org.junit.jupiter.api.AfterEach
import org.junit.jupiter.api.Test
import org.junit.jupiter.api.assertThrows
import kotlin.test.assertEquals

class DefaultMixTypeLogicTest {
    private val mixTypeServiceMock = mockk<MixTypeService>()
    private val materialLogicMock = mockk<MaterialLogic>()

    private val mixTypeLogic = spyk(DefaultMixTypeLogic(mixTypeServiceMock, materialLogicMock))

    private val materialType = MaterialTypeDto(1L, "Unit test material type", "UTMT", false)
    private val material = MaterialDto(1L, "Unit test material", 1000f, true, materialType)
    private val mixType = MixTypeDto(id = 1L, name = "Unit test mix type", material)

    @AfterEach
    fun afterEach() {
        clearAllMocks()
    }

    @Test
    fun getOrCreateForNameAndMaterialType_normalBehavior_returnsFromService() {
        // Arrange
        every { mixTypeServiceMock.getByNameAndMaterialType(any(), any()) } returns mixType

        // Act
        val actualMixType = mixTypeLogic.getOrCreateForNameAndMaterialType(mixType.name, materialType)

        // Assert
        assertEquals(mixType, actualMixType)
    }

    @Test
    fun getOrCreateForNameAndMaterialType_notFound_returnsFromSaveForNameAndMaterialType() {
        // Arrange
        every { mixTypeServiceMock.getByNameAndMaterialType(any(), any()) } returns null
        every { mixTypeLogic.saveForNameAndMaterialType(any(), any()) } returns mixType

        // Act
        val actualMixType = mixTypeLogic.getOrCreateForNameAndMaterialType(mixType.name, materialType)

        // Assert
        assertEquals(mixType, actualMixType)
    }

    @Test
    fun saveForNameAndMaterialType_normalBehavior_callsSavesInMaterialLogic() {
        // Arrange
        every { materialLogicMock.save(any<MaterialDto>()) } returnsArgument 0
        every { mixTypeLogic.save(any()) } returnsArgument 0

        // Act
        mixTypeLogic.saveForNameAndMaterialType(mixType.name, materialType)

        // Assert
        verify {
            materialLogicMock.save(match<MaterialDto> { it.name == mixType.name && it.materialType == materialType })
        }
        confirmVerified(materialLogicMock)
    }

    @Test
    fun saveForNameAndMaterialType_normalBehavior_callsSave() {
        // Arrange
        every { materialLogicMock.save(any<MaterialDto>()) } returnsArgument 0
        every { mixTypeLogic.save(any()) } returnsArgument 0

        // Act
        mixTypeLogic.saveForNameAndMaterialType(mixType.name, materialType)

        // Assert
        verify {
            mixTypeLogic.save(match { it.name == mixType.name && it.material.name == mixType.name && it.material.materialType == materialType })
        }
    }

    @Test
    fun updateForNameAndMaterialType_normalBehavior_callsSavesInMaterialLogic() {
        // Arrange
        val updatedName = mixType.name + " updated"
        val updatedMaterialType = materialType.copy(id = 2L)

        every { materialLogicMock.update(any<MaterialDto>()) } returnsArgument 0
        every { mixTypeLogic.update(any()) } returnsArgument 0

        // Act
        mixTypeLogic.updateForNameAndMaterialType(mixType, updatedName, updatedMaterialType)

        // Assert
        verify {
            materialLogicMock.update(match<MaterialDto> { it.name == updatedName && it.materialType == updatedMaterialType })
        }
        confirmVerified(materialLogicMock)
    }

    @Test
    fun updateForNameAndMaterialType_normalBehavior_callsSave() {
        // Arrange
        val updatedName = mixType.name + " updated"
        val updatedMaterialType = materialType.copy(id = 2L)

        every { materialLogicMock.update(any<MaterialDto>()) } returnsArgument 0
        every { mixTypeLogic.update(any()) } returnsArgument 0

        // Act
        mixTypeLogic.updateForNameAndMaterialType(mixType, updatedName, updatedMaterialType)

        // Assert
        verify {
            mixTypeLogic.update(match { it.name == updatedName && it.material.name == updatedName && it.material.materialType == updatedMaterialType })
        }
    }

    @Test
    fun deleteById_usedByMix_throwsCannotDeleteException() {
        // Arrange
        every { mixTypeServiceMock.isUsedByMixes(any()) } returns true

        // Act
        // Assert
        assertThrows<CannotDeleteException> { mixTypeLogic.deleteById(mixType.id) }
    }
}